"""remotepixel.l8_ndvi"""

import base64
from io import BytesIO

import numpy as np
from PIL import Image

import rasterio as rio
from rasterio import warp
from rasterio.enums import Resampling
from rio_toa import reflectance

from remotepixel import utils

np.seterr(divide='ignore', invalid='ignore')

landsat_bucket = 's3://landsat-pds'


def point(scene, coord):
    """
    """

    scene_params = utils.landsat_parse_scene_id(scene)
    meta_data = utils.landsat_get_mtl(scene)
    landsat_address = f'{landsat_bucket}/{scene_params["key"]}'

    E = float(utils.landsat_mtl_extract(meta_data, 'SUN_ELEVATION'))

    MR = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_MULT_BAND_4'))
    AR = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_ADD_BAND_4'))

    band_address = f'{landsat_address}_B4.TIF'
    with rio.open(band_address) as band:
        lon_srs, lat_srs = warp.transform('EPSG:4326', band.crs, [coord[0]], [coord[1]])
        b4 = list(band.sample([(lon_srs[0], lat_srs[0])]))[0]
        b4 = reflectance.reflectance(b4, MR, AR, E, src_nodata=0)[0]

    MR = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_MULT_BAND_5'))
    AR = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_ADD_BAND_5'))

    band_address = f'{landsat_address}_B5.TIF'
    with rio.open(band_address) as band:
        lon_srs, lat_srs = warp.transform('EPSG:4326', band.crs, [coord[0]], [coord[1]])
        b5 = list(band.sample([(lon_srs[0], lat_srs[0])]))[0]
        b5 = reflectance.reflectance(b5, MR, AR, E, src_nodata=0)[0]

    ratio = np.nan_to_num((b5 - b4) / (b5 + b4)) if (b4 * b5) > 0 else 0.

    out = {
        'ndvi': ratio,
        'date': scene_params['date'],
        'cloud': float(utils.landsat_mtl_extract(meta_data, 'CLOUD_COVER'))}

    return out


def area(scene, bbox):
    """
    """

    max_width = 512
    max_height = 512

    scene_params = utils.landsat_parse_scene_id(scene)
    meta_data = utils.landsat_get_mtl(scene)
    landsat_address = f'{landsat_bucket}/{scene_params["key"]}'

    E = float(utils.landsat_mtl_extract(meta_data, 'SUN_ELEVATION'))

    MR = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_MULT_BAND_4'))
    AR = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_ADD_BAND_4'))

    band_address = f'{landsat_address}_B4.TIF'
    with rio.open(band_address) as band:
        crs_bounds = warp.transform_bounds('EPSG:4326', band.crs, *bbox)
        window = band.window(*crs_bounds, boundless=True)

        width = window.num_cols if window.num_cols < max_width else max_width
        height = window.num_rows if window.num_rows < max_height else max_height

        b4 = band.read(window=window, out_shape=(height, width), indexes=1, resampling=Resampling.bilinear, boundless=True)
        b4 = reflectance.reflectance(b4, MR, AR, E, src_nodata=0)

    MR = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_MULT_BAND_5'))
    AR = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_ADD_BAND_5'))

    band_address = f'{landsat_address}_B5.TIF'
    with rio.open(band_address) as band:
        crs_bounds = warp.transform_bounds('EPSG:4326', band.crs, *bbox)
        window = band.window(*crs_bounds, boundless=True)

        width = window.num_cols if window.num_cols < max_width else max_width
        height = window.num_rows if window.num_rows < max_width else max_width

        b5 = band.read(window=window, out_shape=(height, width), indexes=1, resampling=Resampling.bilinear, boundless=True)
        b5 = reflectance.reflectance(b5, MR, AR, E, src_nodata=0)

    ratio = np.where((b5 * b4) > 0, np.nan_to_num((b5 - b4) / (b5 + b4)), -1)
    ratio = np.where(ratio > -1, utils.linear_rescale(ratio, in_range=[-1, 1], out_range=[1, 255]), 0).astype(np.uint8)

    cmap = list(np.array(utils.get_colormap()).flatten())
    img = Image.fromarray(ratio, 'P')
    img.putpalette(cmap)
    img = img.convert('RGB')

    sio = BytesIO()
    img.save(sio, 'jpeg', subsampling=0, quality=100)
    sio.seek(0)

    return base64.b64encode(sio.getvalue()).decode()



"""remotepixel.l8_full"""

from datetime import datetime, timedelta

import boto3
import numpy as np

import rasterio as rio
from rasterio.io import MemoryFile
from rio_toa import reflectance

from remotepixel import utils

np.seterr(divide='ignore', invalid='ignore')

landsat_bucket = 's3://landsat-pds'


def create(scene, bucket, bands=[4, 3, 2]):
    """
    """

    scene_params = utils.landsat_parse_scene_id(scene)
    meta_data = utils.landsat_get_mtl(scene)
    landsat_address = f'{landsat_bucket}/{scene_params["key"]}'

    bqa = f'{landsat_address}_BQA.TIF'
    with rio.open(bqa) as src:
        meta = src.meta
        wind = [w for ij, w in src.block_windows(1)]

    meta.update(nodata=0,
                count=3,
                interleave='pixel',
                PHOTOMETRIC='RGB',
                tiled=False,
                compress=None)

    E = float(utils.landsat_mtl_extract(meta_data, 'SUN_ELEVATION'))

    with MemoryFile() as memfile:
        with memfile.open(**meta) as dataset:

            for b in range(len(bands)):
                band_address = f'{landsat_address}_B{bands[b]}.TIF'

                MR = float(utils.landsat_mtl_extract(meta_data, f'REFLECTANCE_MULT_BAND_{bands[b]}'))
                AR = float(utils.landsat_mtl_extract(meta_data, f'REFLECTANCE_ADD_BAND_{bands[b]}'))

                with rio.open(band_address) as src:
                    for window in wind:
                        matrix = src.read(window=window, boundless=True, indexes=1)
                        result = reflectance.reflectance(matrix, MR, AR, E, src_nodata=0) * 10000
                        dataset.write(result.astype(np.uint16), window=window, indexes=b+1)

        client = boto3.client('s3')
        str_band = ''.join(map(str, bands))
        key = f'data/landsat/{scene}_B{str_band}.tif'
        expiration = datetime.now() + timedelta(days=15)

        client.put_object(
            ACL='public-read',
            Bucket=bucket,
            Key=key,
            Expires=expiration,
            Body=memfile,
            ContentType='image/tiff')

    return True


def create_ndvi(scene, bucket):
    """
    """

    scene_params = utils.landsat_parse_scene_id(scene)
    meta_data = utils.landsat_get_mtl(scene)
    landsat_address = f'{landsat_bucket}/{scene_params["key"]}'

    bqa = f'{landsat_address}_BQA.TIF'
    with rio.open(bqa) as src:
        meta = src.meta
        wind = [w for ij, w in src.block_windows(1)]

    meta.update(nodata=-1,
                count=1,
                interleave='pixel',
                tiled=False,
                compress=None,
                dtype=np.float32)

    E = float(utils.landsat_mtl_extract(meta_data, 'SUN_ELEVATION'))

    with MemoryFile() as memfile:
        with memfile.open(**meta) as dataset:

            band_address_b = f'{landsat_address}_B4.TIF'
            MR_B = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_MULT_BAND_4'))
            AR_B = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_ADD_BAND_4'))

            band_address_n = f'{landsat_address}_B5.TIF'
            MR_N = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_MULT_BAND_5'))
            AR_N = float(utils.landsat_mtl_extract(meta_data, 'REFLECTANCE_ADD_BAND_5'))

            with rio.open(band_address_b) as b4:
                with rio.open(band_address_n) as b5:
                    for window in wind:
                        matrix = b4.read(window=window, boundless=True, indexes=1)
                        b4_data = reflectance.reflectance(matrix, MR_B, AR_B, E, src_nodata=0)

                        matrix = b5.read(window=window, boundless=True, indexes=1)
                        b5_data = reflectance.reflectance(matrix, MR_N, AR_N, E, src_nodata=0)

                        ratio = np.where((b5_data * b4_data) > 0, np.nan_to_num((b5_data - b4_data) / (b5_data + b4_data)), -1)

                        dataset.write(ratio, window=window, indexes=1)

        client = boto3.client('s3')
        key = f'data/landsat/{scene}_NDVI.tif'
        expiration = datetime.now() + timedelta(days=15)

        client.put_object(
            ACL='public-read',
            Bucket=bucket,
            Key=key,
            Expires=expiration,
            Body=memfile,
            ContentType='image/tiff')

    return True



"""remotepixel.srtm_mosaic.py"""

import os
import zlib
import boto3
from concurrent import futures
from datetime import datetime, timedelta

import numpy as np

import rasterio as rio
from rasterio.merge import merge
from rasterio.io import MemoryFile

from urllib.request import urlopen

srtm_site = 'https://s3.amazonaws.com/elevation-tiles-prod/skadi'


def worker(tile):
    """
    """

    try:
        outpath = f'/tmp/{tile}.hgt'
        url = f'{srtm_site}/{tile[0:3]}/{tile}.hgt.gz'
        with open(outpath, 'wb') as f:
            f.write(zlib.decompress(urlopen(url).read(), zlib.MAX_WBITS | 16))
        return outpath
    except:
        return None


def create(tiles, uuid, bucket):
    """
    """

    if len(tiles) > 8:
        return False

    with futures.ThreadPoolExecutor(max_workers=8) as executor:
        responses = list(executor.map(worker, tiles))

    sources = [rio.open(tile) for tile in responses if tile]
    dest, output_transform = merge(sources, nodata=-32767)

    with MemoryFile() as memfile:
        with memfile.open(driver='GTiff',
                          count=1,
                          dtype=np.int16,
                          nodata=-32767,
                          height=dest.shape[1],
                          width=dest.shape[2],
                          compress='DEFLATE',
                          crs='epsg:4326',
                          transform=output_transform) as dataset:
                            dataset.write(dest)

        client = boto3.client('s3')
        expiration = datetime.now() + timedelta(days=7)

        client.put_object(
            ACL='public-read',
            Bucket=bucket,
            Key=os.path.join('data', 'srtm', '{}.tif'.format(uuid)),
            Expires=expiration,
            Body=memfile,
            ContentType='image/tiff')

    return True



"""remotepixel.l8_mosaic"""

import os
import json
from concurrent import futures
from datetime import datetime, timedelta

import boto3
import numpy as np

import rasterio as rio
from rasterio.merge import merge
from rasterio.io import MemoryFile
from rasterio.vrt import WarpedVRT
from rasterio.enums import Resampling
from rasterio.warp import transform_bounds, calculate_default_transform
from rio_toa import reflectance

from remotepixel import utils

landsat_bucket = 's3://landsat-pds'


def get_scene(args):
    """
    """

    scene, bands = args

    try:
        scene_params = utils.landsat_parse_scene_id(scene)
        meta_data = utils.landsat_get_mtl(scene)
        landsat_address = f'{landsat_bucket}/{scene_params["key"]}'

        bqa = f'{landsat_address}_BQA.TIF'
        with rio.open(bqa) as src:
            ovr = src.overviews(1)
            ovr_width = int(src.width / ovr[0])
            ovr_height = int(src.height / ovr[0])

            dst_affine, width, height = calculate_default_transform(src.crs, 'epsg:3857', ovr_width, ovr_height, *src.bounds)

        outpath = f'/tmp/{scene}.tif'
        with rio.open(outpath,
                      'w',
                      driver='GTiff',
                      count=3,
                      dtype=np.uint8,
                      nodata=0,
                      height=height,
                      width=width,
                      crs='epsg:3857',
                      transform=dst_affine) as dataset:

            for b in range(len(bands)):
                band_address = f'{landsat_address}_B{bands[b]}.TIF'

                with rio.open(band_address) as src:
                    with WarpedVRT(src, dst_crs='EPSG:3857', resampling=Resampling.bilinear, src_nodata=0, dst_nodata=0) as vrt:
                        matrix = vrt.read(indexes=1, out_shape=(height, width))

                MR = float(utils.landsat_mtl_extract(meta_data, f'REFLECTANCE_MULT_BAND_{bands[b]}'))
                AR = float(utils.landsat_mtl_extract(meta_data, f'REFLECTANCE_ADD_BAND_{bands[b]}'))
                E = float(utils.landsat_mtl_extract(meta_data, 'SUN_ELEVATION'))

                matrix = reflectance.reflectance(matrix, MR, AR, E, src_nodata=0) * 10000
                minRef = float(utils.landsat_mtl_extract(meta_data, f'REFLECTANCE_MINIMUM_BAND_{bands[b]}')) * 10000
                maxRef = float(utils.landsat_mtl_extract(meta_data, f'REFLECTANCE_MAXIMUM_BAND_{bands[b]}')) * 10000

                matrix = np.where(matrix > 0, utils.linear_rescale(matrix, in_range=[int(minRef), int(maxRef)], out_range=[1, 255]), 0).astype(np.uint8)

                mask = np.ma.masked_values(matrix, 0)
                s = np.ma.notmasked_contiguous(mask)
                mask = None
                matrix = matrix.ravel()
                for sl in s:
                    matrix[sl.start: sl.start + 5] = 0
                    matrix[sl.stop - 5:sl.stop] = 0
                matrix = matrix.reshape((height, width))

                dataset.write(matrix, indexes=b+1)

        return outpath

    except:
        return None


def create(scenes, uuid, bucket, bands=[4, 3, 2]):
    """
    """

    args = ((scene, bands) for scene in scenes)
    with futures.ThreadPoolExecutor(max_workers=10) as executor:
        allScenes = executor.map(get_scene, args)

    sources = [rio.open(x) for x in allScenes if x]
    dest, output_transform = merge(sources, nodata=0)

    for tmp in allScenes:
        if tmp:
            os.remove(tmp)

    with MemoryFile() as memfile:
        with memfile.open(driver='GTiff',
                          count=3,
                          dtype=np.uint8,
                          nodata=0,
                          height=dest.shape[1],
                          width=dest.shape[2],
                          compress='JPEG',
                          crs='epsg:3857',
                          transform=output_transform) as dataset:

            dataset.write(dest)
            wgs_bounds = transform_bounds(
                *[dataset.crs, 'epsg:4326'] +
                list(dataset.bounds), densify_pts=21)

        client = boto3.client('s3')
        expiration = datetime.now() + timedelta(days=7)

        client.put_object(
            ACL='public-read',
            Bucket=os.environ.get('OUTPUT_BUCKET'),
            Key=f'data/mosaic/{uuid}_mosaic.tif',
            Expires=expiration,
            Body=memfile,
            ContentType='image/tiff')

        meta = {
            'id': uuid,
            'mosaic': '{}_mosaic.tif'.format(uuid),
            'coordinates': {
                'north': wgs_bounds[3],
                'west': wgs_bounds[0],
                'south': wgs_bounds[1],
                'east': wgs_bounds[2],
                'Proj': 'EPSG:4326'}}

        client.put_object(
            ACL='public-read',
            Bucket=bucket,
            Key=f'data/mosaic/{uuid}.json',
            Expires=expiration,
            Body=json.dumps(meta),
            ContentType='application/json')

    return True



import os
import re
import datetime

from urllib.request import urlopen

import numpy as np

import rasterio as rio
from rasterio.enums import Resampling


def get_colormap():
    """
    """
    cmap_file = os.path.join(os.path.dirname(__file__), 'cmap.txt')
    with open(cmap_file) as cmap:
        lines = cmap.read().splitlines()
        colormap = [list(map(int, line.split())) for line in lines if not line.startswith('#')][1:]

    return colormap


def get_overview(address, ovrSize):
    """
    """
    with rio.open(address) as src:
        matrix = src.read(indexes=1, out_shape=(ovrSize, ovrSize), resampling=Resampling.bilinear).astype(src.profile['dtype'])

    return matrix


def linear_rescale(image, in_range=[0, 16000], out_range=[1, 255]):
    """
    Linear rescaling
    """

    imin, imax = in_range
    omin, omax = out_range
    image = np.clip(image, imin, imax) - imin
    image = image / float(imax - imin)

    return (image * (omax - omin) + omin)


def landsat_get_mtl(sceneid):
    """
    Get Landsat MTL metadata
    """

    try:
        scene_params = landsat_parse_scene_id(sceneid)
        meta_file = f'http://landsat-pds.s3.amazonaws.com/{scene_params["key"]}_MTL.txt'
        return urlopen(meta_file).read().decode().splitlines()
    except:
        raise Exception(f'Could not retrieve {sceneid} metadata')


def landsat_mtl_extract(meta, param):
    """
    Extracting MTL info
    """

    for line in meta:
        data = line.split(' = ')
        if (data[0]).strip() == param:
            return (data[1]).strip()

    raise ValueError(f'No {param} found')


def landsat_parse_scene_id(sceneid):
    """
    Author @perrygeo - http://www.perrygeo.com

    parse scene id
    """

    if not re.match('^(L[COTEM]8\d{6}\d{7}[A-Z]{3}\d{2})|(L[COTEM]08_L\d{1}[A-Z]{2}_\d{6}_\d{8}_\d{8}_\d{2}_(T1|T2|RT))$', sceneid):
        raise ValueError(f'Could not match {sceneid}')

    precollection_pattern = (
        r'^L'
        r'(?P<sensor>\w{1})'
        r'(?P<satellite>\w{1})'
        r'(?P<path>[0-9]{3})'
        r'(?P<row>[0-9]{3})'
        r'(?P<acquisitionYear>[0-9]{4})'
        r'(?P<acquisitionJulianDay>[0-9]{3})'
        r'(?P<groundStationIdentifier>\w{3})'
        r'(?P<archiveVersion>[0-9]{2})$')

    collection_pattern = (
        r'^L'
        r'(?P<sensor>\w{1})'
        r'(?P<satellite>\w{2})'
        r'_'
        r'(?P<processingCorrectionLevel>\w{4})'
        r'_'
        r'(?P<path>[0-9]{3})'
        r'(?P<row>[0-9]{3})'
        r'_'
        r'(?P<acquisitionYear>[0-9]{4})'
        r'(?P<acquisitionMonth>[0-9]{2})'
        r'(?P<acquisitionDay>[0-9]{2})'
        r'_'
        r'(?P<processingYear>[0-9]{4})'
        r'(?P<processingMonth>[0-9]{2})'
        r'(?P<processingDay>[0-9]{2})'
        r'_'
        r'(?P<collectionNumber>\w{2})'
        r'_'
        r'(?P<collectionCategory>\w{2})$')

    meta = None
    for pattern in [collection_pattern, precollection_pattern]:
        match = re.match(pattern, sceneid, re.IGNORECASE)
        if match:
            meta = match.groupdict()
            break

    if not meta:
        raise ValueError(f'Could not match {sceneid}')

    if meta.get('acquisitionJulianDay'):
        date = datetime.datetime(int(meta['acquisitionYear']), 1, 1) \
            + datetime.timedelta(int(meta['acquisitionJulianDay']) - 1)

        meta['date'] = date.strftime('%Y-%m-%d')
    else:
        meta['date'] = f'{meta.get("acquisitionYear")}-{meta.get("acquisitionMonth")}-{meta.get("acquisitionDay")}'

    collection = meta.get('collectionNumber', '')
    if collection != '':
        collection = f'c{int(collection)}'

    meta['key'] = os.path.join(collection, 'L8', meta['path'], meta['row'], sceneid, sceneid)
    meta['scene'] = sceneid

    return meta


def sentinel_parse_scene_id(sceneid):
    """
    parse scene id
    """

    if not re.match('^S2[AB]_tile_[0-9]{8}_[0-9]{2}[A-Z]{3}_[0-9]$', sceneid):
        raise ValueError(f'Could not match {sceneid}')

    sentinel_pattern = (
        r'^S'
        r'(?P<sensor>\w{1})'
        r'(?P<satellite>[AB]{1})'
        r'_tile_'
        r'(?P<acquisitionYear>[0-9]{4})'
        r'(?P<acquisitionMonth>[0-9]{2})'
        r'(?P<acquisitionDay>[0-9]{2})'
        r'_'
        r'(?P<utm>[0-9]{2})'
        r'(?P<sq>\w{1})'
        r'(?P<lat>\w{2})'
        r'_'
        r'(?P<num>[0-9]{1})$')

    meta = None
    match = re.match(sentinel_pattern, sceneid, re.IGNORECASE)
    if match:
        meta = match.groupdict()

    if not meta:
        raise ValueError('Could not match {sceneid}')

    utm = meta['utm']
    sq = meta['sq']
    lat = meta['lat']
    year = meta['acquisitionYear']
    m = meta['acquisitionMonth'].lstrip("0")
    d = meta['acquisitionDay'].lstrip("0")
    n = meta['num']

    meta['key'] = f'tiles/{utm}/{sq}/{lat}/{year}/{m}/{d}/{n}'

    return meta




import os

from setuptools import setup, find_packages

with open('remotepixel/__init__.py') as f:
    for line in f:
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"')
            version = version.strip("'")
            break


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='remotepixel',
      version=version,
      description=u"",
      long_description=u"",
      classifiers=['Programming Language :: Python :: 3.6'],
      keywords='remotepixel AWS lambda Landsat Sentinel SRTM',
      author=u"RemotePixel",
      author_email='contact@remotepixel.ca',
      url='https://github.com/remotepixel/remotepixel-py',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      python_requires='~=3.6',
      package_data={'remotepixel': ['cmap.txt']},
      install_requires=[
          "numpy",
          "Pillow",
          "mercantile",
          "rasterio[s3]==1.0a9",
          "rio-toa"
      ],
      extras_require={
          'test': ['pytest', 'pytest-cov', 'codecov'],
      })



"""remotepixel.s2_full"""

from concurrent import futures
from datetime import datetime, timedelta

import boto3
import numpy as np

import rasterio as rio
from rasterio.io import MemoryFile

from remotepixel import utils

sentinel_bucket = 's3://sentinel-s2-l1c'

################################################################################

# This code won't run on AWS Lambda

# TO DO: handle multi resolution

################################################################################

# https://en.wikipedia.org/wiki/Sentinel-2
band_info = {
    '01': {'res': 60, 'wavelenght': 0.443, 'name': 'Coastal aerosol'},
    '02': {'res': 10, 'wavelenght': 0.490, 'name': 'Blue'},
    '03': {'res': 10, 'wavelenght': 0.560, 'name': 'Green'},
    '04': {'res': 10, 'wavelenght': 0.665, 'name': 'Red'},
    '05': {'res': 20, 'wavelenght': 0.705, 'name': 'Vegetation Red Edge'},
    '06': {'res': 20, 'wavelenght': 0.740, 'name': 'Vegetation Red Edge'},
    '07': {'res': 20, 'wavelenght': 0.783, 'name': 'Vegetation Red Edge'},
    '08': {'res': 10, 'wavelenght': 0.842, 'name': 'NIR'},
    '8A': {'res': 20, 'wavelenght': 0.865, 'name': 'Vegetation Red Edge'},
    '09': {'res': 60, 'wavelenght': 0.945, 'name': 'Water vapour'},
    '10': {'res': 60, 'wavelenght': 1.375, 'name': 'SWIR'},
    '11': {'res': 20, 'wavelenght': 1.610, 'name': 'SWIR'},
    '12': {'res': 20, 'wavelenght': 2.190, 'name': 'SWIR'}}


def worker(band_address):
    """
    """
    with rio.open(band_address) as src:
        return src.read(indexes=1)


def create(scene, bucket, bands=['04', '03', '02']):
    """
    """

    scene_params = utils.sentinel_parse_scene_id(scene)
    sentinel_address = f'{sentinel_bucket}/{scene_params["key"]}'

    band_address = f'{sentinel_address}/B{bands[0]}.jp2'
    with rio.open(band_address) as src:
        meta = src.meta

    meta.update(driver='GTiff',
                nodata=0,
                count=3,
                interleave='pixel',
                PHOTOMETRIC='RGB',
                tiled=False,
                compress=None)

    addresses = [f'{sentinel_address}/B{band}.jp2' for band in bands]

    with MemoryFile() as memfile:
        with memfile.open(**meta) as dataset:
            with futures.ThreadPoolExecutor(max_workers=3) as executor:
                dataset.write(np.stack(list(executor.map(worker, addresses))))

        client = boto3.client('s3')
        str_band = ''.join(map(str, bands))
        key = f'data/s2/{scene}_B{str_band}.tif'
        expiration = datetime.now() + timedelta(days=15)

        client.put_object(
            ACL='public-read',
            Bucket=bucket,
            Key=key,
            Expires=expiration,
            Body=memfile,
            ContentType='image/tiff')

    return True



"""remotepixel.l8_ovr"""

import base64
from io import BytesIO
from concurrent import futures

import numpy as np
from PIL import Image

from rio_toa import reflectance

from remotepixel import utils

np.seterr(divide='ignore', invalid='ignore')

landsat_bucket = 's3://landsat-pds'


def worker(args):
    """
    """

    address, band, meta, ovrSize, ndvi = args

    MR = float(utils.landsat_mtl_extract(meta, f'REFLECTANCE_MULT_BAND_{band}'))
    AR = float(utils.landsat_mtl_extract(meta, f'REFLECTANCE_ADD_BAND_{band}'))
    E = float(utils.landsat_mtl_extract(meta, 'SUN_ELEVATION'))

    matrix = utils.get_overview(address, ovrSize)
    matrix = reflectance.reflectance(matrix, MR, AR, E, src_nodata=0)
    if not ndvi:
        imgRange = np.percentile(matrix[matrix > 0], (2, 98)).tolist()
        matrix = np.where(matrix > 0, utils.linear_rescale(matrix, in_range=imgRange, out_range=[1, 255]), 0).astype(np.uint8)

    return matrix


def create(scene, bands=[4, 3, 2], img_format='jpeg', ovrSize=512):
    """
    """

    if img_format not in ['png', 'jpeg']:
        raise UserWarning(f'Invalid {img_format} extension')

    scene_params = utils.landsat_parse_scene_id(scene)
    meta_data = utils.landsat_get_mtl(scene)
    landsat_address = f'{landsat_bucket}/{scene_params["key"]}'

    args = ((f'{landsat_address}_B{band}.TIF',  band, meta_data, ovrSize, False) for band in bands)

    out = np.zeros((4, ovrSize, ovrSize), dtype=np.uint8)
    with futures.ThreadPoolExecutor(max_workers=3) as executor:
        out[0:3] = list(executor.map(worker, args))

    out[-1] = np.all(np.dstack(out[:3]) != 0, axis=2).astype(np.uint8) * 255

    img = Image.fromarray(np.dstack(out))
    sio = BytesIO()

    if img_format == 'jpeg':
        img = img.convert('RGB')
        img.save(sio, 'jpeg', subsampling=0, quality=100)
    else:
        img.save(sio, 'png', compress_level=0)

    sio.seek(0)

    return base64.b64encode(sio.getvalue()).decode()


def create_ndvi(scene, img_format='jpeg', ovrSize=512):
    """
    """

    if img_format not in ['png', 'jpeg']:
        raise UserWarning(f'Invalid {img_format} extension')

    scene_params = utils.landsat_parse_scene_id(scene)
    meta_data = utils.landsat_get_mtl(scene)
    landsat_address = f'{landsat_bucket}/{scene_params["key"]}'

    bands = [4, 5]

    args = ((f'{landsat_address}_B{band}.TIF', band, meta_data, ovrSize, True) for band in bands)

    with futures.ThreadPoolExecutor(max_workers=3) as executor:
        out = list(executor.map(worker, args))

    ratio = np.where((out[1] * out[0]) > 0, np.nan_to_num((out[1] - out[0]) / (out[1] + out[0])), -1)
    ratio = np.where(ratio > -1, utils.linear_rescale(ratio, in_range=[-1, 1], out_range=[1, 255]), 0).astype(np.uint8)

    cmap = list(np.array(utils.get_colormap()).flatten())
    img = Image.fromarray(ratio, 'P')
    img.putpalette(cmap)

    sio = BytesIO()
    if img_format == 'jpeg':
        img = img.convert('RGB')
        img.save(sio, 'jpeg', subsampling=0, quality=100)
    else:
        img.save(sio, 'png', compress_level=0)

    sio.seek(0)

    return base64.b64encode(sio.getvalue()).decode()
